import torch
import sys
from torch.utils.data import TensorDataset

from EIVGeneral.manipulate_tensors import add_noise, normalize_tensor,\
        unnormalize_tensor

total_number_of_datapoints = 500
input_range = [-1,1]
slope = 1.0
intercept = 0.0
x_noise_strength = 0.1
y_noise_strength = 0.1
func = lambda true_x: slope * true_x**2 + intercept 
def design_matrix(x, device):
    x = x.to(device)
    assert len(x.shape) == 2 and x.shape[1] == 1
    return torch.cat((torch.ones(x.shape).to(device), x, x**2), dim=1)



def load_data(seed=0, splitting_part=0.8, normalize=True,
        return_ground_truth=False,
        return_normalized_func=False,
        fixed_seed = 0):
    """
    Loads one-dimensional data
    :param seed: Seed for shuffling the data before splitting.
    :param splitting_part: Which fraction of the data to use as training
    data. Defaults to 0.8.
    :param normalize: Whether to normalize the data, defaults to True.
    :param return_ground_truth: Boolean. If True, the unnoisy ground truth will
    also be returned. Defaults to False.
    :param return_normalized_func: Boolean (default False). If True, the
    normalized version of the used function is returned as a last element.
    :param fixed_seed: Used to generate the full dataset (test and train).
    Defaults to 0.
    :returns: quadratic_trainset, quadratic_testset, (, normalized_func) if
    return_ground_truth is False,
    else quadratic_trainset, quadratic_testset,  true_quadratic_trainset,
    true_quadratic_testset, (, normalized_func). The "true" datasets each return
    **four tensors**: The true x,y and their noisy counterparts.
    """
    # draw different seeds for noise and splitting
    random_generator = torch.Generator().manual_seed(fixed_seed)
    seeds = [int(t) for t in torch.randint(0,sys.maxsize,(3,),\
            generator=random_generator)]

    # create new generators from tensor seeds
    true_x = input_range[0] + (input_range[1]-input_range[0])\
                  * torch.rand((total_number_of_datapoints,1),
                          generator=torch.Generator().manual_seed(seeds[0]))
    true_y = func(true_x)

    # add noise and normalize x and y
    (noisy_x, noisy_y), (true_x, true_y), normalization_list = add_noise(
            tensor_list=(true_x, true_y),
            noise_strength_list=(x_noise_strength, y_noise_strength),
            seed_list=seeds[1:3],
            normalize=normalize,
            return_normalization=True)
    if normalize:
        def normalized_func(x):
            unnormalized_x = unnormalize_tensor(x, normalization_list[0])
            y = func(unnormalized_x)
            normalized_y = normalize_tensor(y, normalization_list[1])
            return normalized_y
    else:
        def normalized_func(x):
            return func(x)
    dataset_len = noisy_x.shape[0]

    # shuffle via seed
    new_order = torch.randperm(dataset_len,
            generator=torch.Generator().manual_seed(seed))
    true_x = true_x[new_order, ...]
    true_y = true_y[new_order, ...]
    noisy_x = noisy_x[new_order, ...]
    noisy_y = noisy_y[new_order, ...]

    # create datasets
    train_len = int(dataset_len*splitting_part)
    test_len = dataset_len - train_len
    true_train_x, true_test_x = torch.split(true_x, [train_len, test_len])
    true_train_y, true_test_y = torch.split(true_y, [train_len, test_len])
    noisy_train_x, noisy_test_x = torch.split(noisy_x, [train_len, test_len])
    noisy_train_y, noisy_test_y = torch.split(noisy_y, [train_len, test_len])
    quadratic_trainset = TensorDataset(noisy_train_x, noisy_train_y)
    quadratic_testset = TensorDataset(noisy_test_x, noisy_test_y)
    true_quadratic_trainset = TensorDataset(true_train_x, true_train_y,
            noisy_train_x, noisy_train_y)
    true_quadratic_testset = TensorDataset(true_test_x, true_test_y,
            noisy_test_x, noisy_test_y)

    # return different objects, depending on Booleans
    if not return_ground_truth:
        if not return_normalized_func:
            return quadratic_trainset, quadratic_testset
        else:
            return quadratic_trainset, quadratic_testset, normalized_func
    else:
        if not return_normalized_func:
            return quadratic_trainset, quadratic_testset, true_quadratic_trainset,\
                true_quadratic_testset
        else: 
            return quadratic_trainset, quadratic_testset, true_quadratic_trainset,\
                true_quadratic_testset, normalized_func
