"""
Collect coverages for various coverage factors, networks and dataloaders.
"""
import numpy as np
import torch
import torch.backends.cudnn
from EIVGeneral.coverage_metrics import epistemic_coverage

def get_coverages(not_averaged_predictions, y,\
        q_range=np.linspace(0.1,0.9,num=30),
        noisy_y=True):
    """
    Compute the numerical and theoretical coverage for various coverage
    factors, computed from the `q` in `q_range`, using
    `EIVGeneral.epistemic_coverage` 
    :param not_averaged_predictions: A tuple of tensors as in the output of
    `FNNEIV.predict` with `take_average_of_prediction` set to `False`, i.e.:
    the predictions of the neural net not averaged over the first dimension 
    (the repetition dimension in `FNNEIV.predict`) and
    the aleatoric uncertainty with a batch dimension and a feature dimension. 
    :param y: A `torch.tensor` of the same shape then the second components
    of `not_averaged_predictions`. If the feature dimension is missing, it is added.
    :param q_range: An iterator through floats between 0 and 1.
    :param noisy_y: Boolean. If True (the default), `y` is treated as noisy and
    the total uncertainty is considered. If False, `y` is treated as the
    unnoisy ground truth.
    :returns: numerical_coverage, theoretical_coverage (numpy arrays)
    """
    numerical_coverage_collection = []
    theoretical_coverage_collection = []
    for q in q_range:
        numerical_coverage, theoretical_coverage =\
                epistemic_coverage(\
                not_averaged_predictions=not_averaged_predictions,
                y=y, q=q, noisy_y=noisy_y)
        numerical_coverage_collection.append(float(numerical_coverage))
        theoretical_coverage_collection.append(float(theoretical_coverage))
    return np.array(numerical_coverage_collection),\
            np.array(theoretical_coverage_collection)
    

def get_coverage_distribution(net_iterator, dataloader_iterator,
        device, number_of_draws, q_range=np.linspace(0.1,0.9,num=30),
        number_of_test_samples=10, stack=True,
        noisy_y=True):
    """
    Returns the numerical and theoretical coverage for all nets in
    `net_iterator` with dataloader from `dataloader_iterator`
    :param net_iterator: Iterator yielding `torch.nn.Module`
    :param dataloader_iterator: Iterator yielding `torch.utils.data.DataLoader`.
    Should return the same number of elements than `net_iterator`.
    :param device: The `torch.device` to be used.
    :param number_of_draws: Integer, to be used as keyword for all networks
    returned by `net_iterator`. Defaults to 10.
    :param q_range: Range of floats between 0 and 1, as in 
    `EIVGeneral.coverage_plot.get_coverages`.
    :param number_of_test_samples: An integer, number of samples from the
    dataloaders in dataloader_iterator to be used.
    :param stack: Boolean. If True (default) the results will be stacked along 
    the last dimension. If False a list will be returned.
    :param noisy_y: Boolean. If True (the default), `y` is treated as noisy and
    the total uncertainty is considered. If False, `y` is treated as the
    unnoisy ground truth.
    :returns: num_coverage_collection, th_coverage_collection
    """
    num_coverage_collection, th_coverage_collection = [],[]
    for net, dataloader in\
     zip(net_iterator, dataloader_iterator):
                not_av_pred_collection_out, not_av_pred_collection_sigma,\
                    y_collection = [], [], []
                for i, (x,y) in enumerate(dataloader):
                    if i>= number_of_test_samples:
                        break
                    x, y = x.to(device), y.to(device)
                    not_averaged_predictions = net.predict(x,
                            take_average_of_prediction=False,
                            number_of_draws=number_of_draws)
                    not_av_pred_collection_out.append(
                            not_averaged_predictions[0])
                    not_av_pred_collection_sigma.append(
                            not_averaged_predictions[1])
                    y_collection.append(y)
                not_av_pred_collection = [
                        torch.concat(not_av_pred_collection_out, dim=0), 
                        torch.concat(not_av_pred_collection_sigma, dim=0)]
                y_collection = torch.concat(y_collection, dim=0)
                numerical_coverage, theoretical_coverage = get_coverages(
                        not_averaged_predictions=not_av_pred_collection,
                        y=y_collection, q_range=q_range,
                        noisy_y=noisy_y)
                num_coverage_collection.append(numerical_coverage)
                th_coverage_collection.append(theoretical_coverage)
    if stack:
        num_coverage_collection = np.stack(num_coverage_collection, axis=-1)
        th_coverage_collection = np.stack(th_coverage_collection, axis=-1)
    return num_coverage_collection, th_coverage_collection
