"""
Plot predictions with uncertainties for (simulated) datasets with a ground 
truth. At the moment it is assumed that the used datasets have an output
dimension equal to 1. Plots are only produced for datasets with input dimension
and output dimension 1.
"""
import importlib
import os
import json

import torch

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from EIVArchitectures import Networks
from EIVTrainingRoutines import train_and_store


font = {'family' : 'DejaVu Sans',
        'weight' : 'normal',
        'size'   : 16}

matplotlib.rc('font', **font)
linewidth = 2.0

# coverage factor
k = 1.96

def compute_predictions_and_uncertainties(data, x_range, eiv, number_of_draws,
        x_noise_seed = 0, y_noise_seed = 1,
        plotting_seed = 0, number_of_test_datapoints = 100):
    """
    Iterate through networks associated with `data` (short dataname) 
    throuth their JSON configuration and evaluate their prediction and
    uncertainties for the (true) x in `x_range`. The results returned are as numpy
    arrays included in a `plotting_dictionary` that contains the predictions
    and uncertainties via the keys "prediction" and "uncertainty" but also 
    - A tuple of `x_range` and the corresponding values of y (key
            "range_points")
    - the noisified version of `x_range` and the corresponding y values (key
    "noisy_range_points") and 
    - `number_of_test_datapoints` points from the test
    dataset with seed `plotting_seed` (key "test_data_points") and
    - the input dimension (key "input_dim").

    **Note**: The output of the neural networks are assumed to be
    one-dimensional .

    :data: String, short dataname. The corresponding module should contain
    `x_noise_strength`, `y_noise_strength`.
    :x_range: An iterator yielding the (true) x values to consider
    :eiv: Boolean. If True an EiV model is used, else an non-EiV model.
    :number_of_draws: Number of draws to use for prediction. Take an int for
    non-EiV models and a two-element list for EiV models.
    :x_noise_seed: An integer. Will be used as a seed to generate the noise put
    on `x_range`.
    :y_noise_seed: An integer. Will be used as a seed to generate the noise put
    on `normalized_func(x_range)` that will be returned with `range_values` in the
    `plotting_dictionary`.
    :plotting_seed: An integer. Needed for choosing which of the test datasets
    will be used to returning the test datapoints.
    :number_of_test_datapoints: A positive integer. Number of test_datapoints
    :returns: plotting_dictionary
    """
    plotting_dictionary = {}
    # load configuration file
    if eiv:
        with open(os.path.join('configurations',f'eiv_{data}.json'),'r') as\
                conf_file:
            conf_dict = json.load(conf_file)
    else:
        with open(os.path.join('configurations',f'noneiv_{data}.json'),'r') as\
                conf_file:
            conf_dict = json.load(conf_file)

    # get datanames
    long_dataname = conf_dict["long_dataname"]
    short_dataname = conf_dict["short_dataname"]
    try:
        normalize = conf_dict['normalize']
    except KeyError:
        # normalize by default
        normalize = True


    # load hyperparameters
    load_data = importlib.import_module(f'EIVData.{long_dataname}').load_data
    x_noise_strength =\
        importlib.import_module(f'EIVData.{long_dataname}').x_noise_strength
    y_noise_strength =\
        importlib.import_module(f'EIVData.{long_dataname}').y_noise_strength
    seed_list = range(conf_dict["seed_range"][0],
            conf_dict["seed_range"][1])

    # switch to gpu, if possible
    try:
        gpu_number = conf_dict["gpu_number"]
        device = torch.device(f'cuda:{gpu_number}')
        try:
            torch.tensor([0.0]).to(device)
        except RuntimeError:
            if torch.cuda.is_available():
                print('Switched to GPU 0')
                device = torch.device('cuda:0')
            else:
                print('No cuda available, using CPU')
                device = torch.device('cpu')
    except KeyError:
        device = torch.device('cpu')


    # determine dimensions
    _, test_data, normalized_func = load_data(seed=plotting_seed,
            return_ground_truth=False,
            return_normalized_func=True,
            normalize=normalize)
    plotting_dictionary['func'] = normalized_func
    input_dim = test_data[0][0].numel()
    output_dim = test_data[0][1].numel()
    assert output_dim == 1
    # store in plotting_dictionary
    plotting_dictionary['input_dim'] = input_dim

    # store test datapoints
    test_x, test_y = test_data[:number_of_test_datapoints]
    plotting_dictionary['test_data_points'] = (test_x.detach().cpu().numpy(),
            test_y.detach().cpu().numpy())


    # iterator for looping through networks
    def net_iterator(eiv=eiv, seed_list=seed_list):
        """
        Yields EiV models (if `eiv`) or
        non-EiV models (if not `eiv`) for the seeds in
        `seed_list` and `data`.
        """
        if eiv:
            # load parameters
            init_std_y = conf_dict["init_std_y_list"][0]
            unscaled_reg = conf_dict["unscaled_reg"]
            p = conf_dict["p"]
            hidden_layers = conf_dict["hidden_layers"]
            fixed_std_x = conf_dict["fixed_std_x"]
            net = Networks.FNNEIV(p=p, init_std_y=init_std_y,
                    h=[input_dim, *hidden_layers, output_dim],
                    fixed_std_x=fixed_std_x).to(device)
            for seed in seed_list:
                # load network paramaters
                saved_file = os.path.join('saved_networks',
                        f'eiv_{short_dataname}'\
                                f'_init_std_y_{init_std_y:.3f}'\
                                f'_ureg_{unscaled_reg:.1f}'\
                                f'_p_{p:.2f}_fixed_std_x_{fixed_std_x:.3f}'\
                                f'_seed_{seed}.pkl')
                train_and_store.open_stored_training(saved_file=saved_file,
                        net=net, device=device)
                yield net
        else:
            # load parameters
            init_std_y = conf_dict["init_std_y_list"][0]
            unscaled_reg = conf_dict["unscaled_reg"]
            p = conf_dict["p"]
            hidden_layers = conf_dict["hidden_layers"]
            net = Networks.FNNBer(p=p, init_std_y=init_std_y,
                    h=[input_dim, *hidden_layers, output_dim]).to(device)
            for seed in seed_list:
                saved_file = os.path.join('saved_networks',
                            f'noneiv_{short_dataname}'\
                                    f'_init_std_y_{init_std_y:.3f}_ureg_{unscaled_reg:.1f}'\
                                    f'_p_{p:.2f}_seed_{seed}.pkl')
                # load network paramaters
                train_and_store.open_stored_training(saved_file=saved_file,
                        net=net, device=device)
                yield net
    
    # add feature dimension (if necessary) 
    x_range = x_range.view((-1, input_dim))
    y_range = normalized_func(x_range)
    # noisify
    noisy_x_range = x_range + x_noise_strength * torch.randn(x_range.shape,
            generator=torch.Generator().manual_seed(x_noise_seed))
    # move to device for later processing
    noisy_x_range = noisy_x_range.to(device) 
    # y values for noisy_x_range (not on device)
    noisy_y_range = y_range + y_noise_strength *\
    torch.randn(x_range.shape,
            generator=torch.Generator().manual_seed(y_noise_seed))
    # save in plotting_dictionary
    plotting_dictionary['noisy_range_points'] =\
            (noisy_x_range.detach().cpu().numpy(),
            noisy_y_range.detach().cpu().numpy())
    plotting_dictionary['range_points'] =\
            (x_range.detach().cpu().numpy(),
            y_range.detach().cpu().numpy())


    # loop through networks and predict
    mean_collection, unc_collection = [], []
    for net in net_iterator(eiv=eiv):
        mean, unc = net.predict_mean_and_unc(noisy_x_range,
                number_of_draws=number_of_draws)[:2]
        if len(mean.shape) > 1:
                # assume only batch and output dimension
                assert len(mean.shape) == 2
                assert len(unc.shape) == 2
                # assume one-dimensional outputs
                assert mean.shape[1] == 1
                assert unc.shape[1] == 1
                # make one-dimensional
        mean = mean.view((-1,))
        unc = unc.view((-1,))
        mean_collection.append(mean)
        unc_collection.append(unc)

    # stack collections along a new, first dimension
    mean_collection = torch.stack(mean_collection, dim=0)
    unc_collection = torch.stack(unc_collection, dim=0)

    # save average in plotting_dictionary
    plotting_dictionary['prediction'] =\
            torch.mean(mean_collection, dim=0).detach().cpu().numpy()
    plotting_dictionary['uncertainty'] =\
            torch.mean(unc_collection, dim=0).detach().cpu().numpy()
    return plotting_dictionary


data_list = ['linear','quadratic','cubic','sine'] # short datanames
list_x_range = [torch.linspace(-1.0,1.0, 50),
                torch.linspace(-1.0,1.0, 50),
                torch.linspace(-1.0,1.0, 50),
        torch.linspace(-0.2,0.8, 50)]
list_color = [('red','blue')] * len(data_list)
list_number_of_draws = [((100,5), 100)] * len(data_list)

# create an extra zoom plot for zoom_example
zoom_example = 'linear'
# where to zoom
zoom_point = 14
# size of the zoom plot
x_zoom_radius = 0.4
y_zoom_radius = x_zoom_radius

if __name__ == '__main__':
    fignum = 0
    for data, x_range, color, number_of_draws in zip(data_list,
            list_x_range, list_color, list_number_of_draws):
        fignum += 1
        eiv_plotting_dictionary = compute_predictions_and_uncertainties(
                data=data,
                x_range=x_range,
                eiv=True,
                number_of_draws=number_of_draws[0])
        noneiv_plotting_dictionary = compute_predictions_and_uncertainties(
                data=data,
                x_range=x_range,
                eiv=False,
                number_of_draws=number_of_draws[1])
        input_dim = eiv_plotting_dictionary['input_dim']
        if input_dim == 1:
            plt.figure(fignum)
            plt.clf()
            x_values, y_values = eiv_plotting_dictionary['range_points']
            noisy_x_values, _ = eiv_plotting_dictionary['noisy_range_points']
            plt.plot(x_values.flatten(), y_values.flatten(),'-', color='k', linewidth=linewidth)
            eiv_pred = eiv_plotting_dictionary['prediction']
            eiv_unc = eiv_plotting_dictionary['uncertainty']
            plt.plot(x_values, eiv_pred,'-',
                    color=color[0], linewidth=linewidth)
            plt.fill_between(x_values.flatten(), eiv_pred-k * eiv_unc,
                    eiv_pred + k * eiv_unc,
                    color=color[0], alpha=0.5)
            noneiv_pred = noneiv_plotting_dictionary['prediction']
            noneiv_unc = noneiv_plotting_dictionary['uncertainty']
            plt.plot(x_values.flatten(), noneiv_pred,'-',
                    color=color[1], linewidth=linewidth)
            plt.fill_between(x_values.flatten(), noneiv_pred-k * noneiv_unc,
                    noneiv_pred + k * noneiv_unc,
                    color=color[1], alpha=0.5)
            plt.tight_layout()
            plt.savefig(f'results/figures/prediction_{data}.pdf')
            if data == zoom_example:
                fignum += 1
                func = eiv_plotting_dictionary['func'] 
                x_point = x_values[zoom_point] 
                y_point = func(torch.tensor(x_point)).numpy()
                noisy_x_point = noisy_x_values[zoom_point]
                func_noisy_x_point = func(torch.tensor(noisy_x_point)).numpy()
                plt.figure(fignum)
                plt.clf()
                plt.plot(x_values.flatten(), y_values.flatten(),'-', color='k', linewidth=linewidth)
                plt.plot(x_values, eiv_pred,'-',
                        color=color[0], linewidth=linewidth)
                plt.fill_between(x_values.flatten(), eiv_pred-k * eiv_unc,
                        eiv_pred + k * eiv_unc,
                        color=color[0], alpha=0.5)
                plt.plot(x_values.flatten(), noneiv_pred,'-',
                        color=color[1], linewidth=linewidth)
                plt.fill_between(x_values.flatten(), noneiv_pred-k * noneiv_unc,
                        noneiv_pred + k * noneiv_unc,
                        color=color[1], alpha=0.5)
                plt.axvline(x_point, color='black', linestyle='dotted')
                plt.axhline(y_point, color='black', linestyle='dotted')
                plt.axvline(noisy_x_point, color='gray', linestyle='dashed')
                plt.axhline(func_noisy_x_point, color='gray', linestyle='dashed')
                plt.text(x_point - 0.1 * x_zoom_radius,y_point-0.9 * y_zoom_radius, r'$\zeta$', color='k')
                plt.text(noisy_x_point - 0.1 * x_zoom_radius,y_point-0.9 * y_zoom_radius, r'$x$', color='gray')
                plt.text(x_point - 0.92 * x_zoom_radius,y_point-0.13 * y_zoom_radius, r'$g(\zeta)$', color='k')
                plt.text(x_point - 0.92 * x_zoom_radius,func_noisy_x_point-0.13 * y_zoom_radius, r'$g(x)$', color='gray')
                plt.gca().set_xlim(left=x_point - x_zoom_radius, 
                        right=x_point + x_zoom_radius)
                plt.gca().set_ylim(bottom=y_point - y_zoom_radius,
                        top=y_point + y_zoom_radius)
                plt.gca().set_aspect('equal', adjustable='box')
                plt.tight_layout()
                plt.savefig(f'results/figures/prediction_{data}_zoom.pdf')
        else:
            # multidimensional handling not included yet
            pass


    plt.show()
