"""
Plot summary quantities in bar plots, that is
- the RMSE (w.r.t. noisy data)
- the total coverage
- the learned std_y
by reading the results produced by `evaluate_metrics.py`
"""
import os
import glob
import json

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

font = {'family' : 'DejaVu Sans',
        'weight' : 'normal',
        'size'   : 16}

matplotlib.rc('font', **font)
k = 1


# load in all available result files
list_of_result_files = glob.glob(os.path.join('results','*.json'))
results = {}
for filename in list_of_result_files:
    data = filename.replace(os.path.join('results','metrics_'),'').replace('.json','')
    with open(filename,'r') as f:
       results[data] = json.load(f)

def save_readout(dictionary, key):
    """
    Returns the value of the `dictionary` for `key`, unless
    the later doesn't exist, in which case (None,None) is returned.
    """
    try:
        readout = dictionary[key]
        if type(readout) is list:
            assert len(readout) == 2
            return readout
        else:
            readout = float(readout)
            return (readout, None)
    except KeyError:
        return (None,None)



## RMSE plot

metric = 'rmse'
data_list = results.keys()
colors = ['red', 'blue']
ymax = 0.8
minimal_bar_size = ymax * 1.5e-3
# read out EiV and non-EiV results for all datasets
metric_results = [
        (save_readout(results[data]['eiv'], metric),
        save_readout(results[data]['noneiv'], metric))
            for data in data_list]

# create figure
plt.figure(1)
plt.clf()
plt.gcf().canvas.manager.set_window_title('RMSE')

# plot bars
for i, ([(eiv_metric_mean, eiv_metric_std),
        (noneiv_metric_mean, noneiv_metric_std)],\
                data) in\
                enumerate(zip(metric_results, data_list)):
    if eiv_metric_mean is not None:
        assert noneiv_metric_mean is not None
        if eiv_metric_std is not None:
            assert noneiv_metric_std is not None
            eiv_bar_size = max(eiv_metric_std, minimal_bar_size)
            noneiv_bar_size = max(noneiv_metric_std, minimal_bar_size)
            plt.plot(i+1, eiv_metric_mean, '^', color=colors[0], markersize=16)
            plt.bar(i+1,
                    height = 2*eiv_bar_size,
                    width = 1.0,
                    bottom = eiv_metric_mean - eiv_bar_size,
                    color=colors[0],
                    alpha=0.5)
            plt.plot(i+1, noneiv_metric_mean, '^', color=colors[1], markersize=16, zorder=0)
            plt.bar(i+1,
                    height = 2 * k *noneiv_bar_size,
                    width = 1.0,
                    bottom = noneiv_metric_mean - k*  noneiv_bar_size,
                    color=colors[1],
                    alpha=0.5)
plt.ylim(bottom=0, top=ymax)
ax = plt.gca()
ax.set_xticks(np.arange(1,len(data_list)+1))
ax.set_xticklabels(data_list, rotation='vertical')
plt.ylabel('RMSE')
plt.tight_layout()
plt.savefig('results/figures/RMSE_bar_plot.pdf')

## coverage plot

metric = 'true_coverage_numerical'
data_list = ['linear','quadratic','cubic','sine']
colors = ['red', 'blue','black']
ymax = 1.0
minimal_bar_size = ymax * 1.5e-3
# read out EiV and non-EiV results for all datasets
metric_results = [
        (save_readout(results[data]['eiv'], metric),
        save_readout(results[data]['noneiv'], metric),
        save_readout(results[data]['lin'], metric))
            for data in data_list]

# create figure
plt.figure(2)
plt.clf()
plt.gcf().canvas.manager.set_window_title('coverage (ground truth)')

# plot bars
for i, ([(eiv_metric_mean, eiv_metric_std),
        (noneiv_metric_mean, noneiv_metric_std),
            (lin_metric_mean, lin_metric_std)],\
                data) in\
                enumerate(zip(metric_results, data_list)):
    if eiv_metric_mean is not None:
        assert noneiv_metric_mean is not None
        if eiv_metric_std is not None:
            assert noneiv_metric_std is not None
            eiv_bar_size = max(eiv_metric_std, minimal_bar_size)
            noneiv_bar_size = max(noneiv_metric_std, minimal_bar_size)
            lin_bar_size = max(lin_metric_std, minimal_bar_size)
            plt.plot(i+1, eiv_metric_mean, '^', color=colors[0], markersize=16)
            plt.bar(i+1,
                    height = 2*eiv_bar_size,
                    width = 0.3,
                    bottom = eiv_metric_mean - eiv_bar_size,
                    color=colors[0],
                    alpha=0.5)
            plt.plot(i+1, noneiv_metric_mean, '^', color=colors[1], markersize=16, zorder=0)
            plt.bar(i+1,
                    height = 2 * k *noneiv_bar_size,
                    width = 0.3,
                    bottom = noneiv_metric_mean - k*  noneiv_bar_size,
                    color=colors[1],
                    alpha=0.5)
            plt.plot(i+1, lin_metric_mean, '^', color=colors[2], markersize=16)
            plt.bar(i+1,
                    height = 2 * k *lin_bar_size,
                    width = 0.3,
                    bottom = lin_metric_mean - k*  lin_bar_size,
                    color=colors[2],
                    alpha=0.5)
plt.axhline(0.95,0.0,1.0,color='k', linestyle='dashed')
plt.ylim(bottom=-0.045, top=ymax)
ax = plt.gca()
ax.set_xticks(np.arange(1,len(data_list)+1))
ax.set_xticklabels(data_list, rotation='vertical')
plt.ylabel('coverage ground truth')
plt.tight_layout()
plt.savefig('results/figures/true_coverage_bar_plot.pdf')

## noisy coverage plot

metric = 'coverage_numerical'
data_list = results.keys()
colors = ['red', 'blue']
ymax = 1.0
minimal_bar_size = ymax * 1.5e-3
# read out EiV and non-EiV results for all datasets
metric_results = [
        (save_readout(results[data]['eiv'], metric),
        save_readout(results[data]['noneiv'], metric))
            for data in data_list]

# create figure
plt.figure(3)
plt.clf()
plt.gcf().canvas.manager.set_window_title('coverage (noisy labels)')

# plot bars
for i, ([(eiv_metric_mean, eiv_metric_std),
        (noneiv_metric_mean, noneiv_metric_std)],\
                data) in\
                enumerate(zip(metric_results, data_list)):
    if eiv_metric_mean is not None:
        assert noneiv_metric_mean is not None
        if eiv_metric_std is not None:
            assert noneiv_metric_std is not None
            eiv_bar_size = max(eiv_metric_std, minimal_bar_size)
            noneiv_bar_size = max(noneiv_metric_std, minimal_bar_size)
            plt.plot(i+1, eiv_metric_mean, '^', color=colors[0], markersize=16)
            plt.bar(i+1,
                    height = 2*eiv_bar_size,
                    width = 0.95,
                    bottom = eiv_metric_mean - eiv_bar_size,
                    color=colors[0],
                    alpha=0.5)
            plt.plot(i+1, noneiv_metric_mean, '^', color=colors[1], markersize=16, zorder=0)
            plt.bar(i+1,
                    height = 2 * k *noneiv_bar_size,
                    width = 0.95,
                    bottom = noneiv_metric_mean - k*  noneiv_bar_size,
                    color=colors[1],
                    alpha=0.5)
plt.ylim(bottom=0, top=ymax)
ax = plt.gca()
ax.set_xticks(np.arange(1,len(data_list)+1))
ax.set_xticklabels(data_list, rotation='vertical')
plt.ylabel('coverage (epis. unc. / ' + r'$u(x)$' + ')' )
plt.tight_layout()
plt.savefig('results/figures/noisy_coverage_bar_plot.pdf')

## coverage by total uncertainty

metric = 'total_coverage'
data_list = results.keys()
colors = ['red', 'blue']
ymax = 1.0
minimal_bar_size = ymax * 1.5e-3
# read out EiV and non-EiV results for all datasets
metric_results = [
        (save_readout(results[data]['eiv'], metric),
        save_readout(results[data]['noneiv'], metric))
            for data in data_list]

# create figure
plt.figure(4)
plt.clf()
plt.gcf().canvas.manager.set_window_title('total_coverage')

# plot bars
for i, ([(eiv_metric_mean, eiv_metric_std),
        (noneiv_metric_mean, noneiv_metric_std)],\
                data) in\
                enumerate(zip(metric_results, data_list)):
    if eiv_metric_mean is not None:
        assert noneiv_metric_mean is not None
        if eiv_metric_std is not None:
            assert noneiv_metric_std is not None
            eiv_bar_size = max(eiv_metric_std, minimal_bar_size)
            noneiv_bar_size = max(noneiv_metric_std, minimal_bar_size)
            plt.plot(i+1, eiv_metric_mean, '^', color=colors[0], markersize=16)
            plt.bar(i+1,
                    height = 2*eiv_bar_size,
                    width = 0.95,
                    bottom = eiv_metric_mean - eiv_bar_size,
                    color=colors[0],
                    alpha=0.5)
            plt.plot(i+1, noneiv_metric_mean, '^', color=colors[1], markersize=16, zorder=0)
            plt.bar(i+1,
                    height = 2 * k *noneiv_bar_size,
                    width = 0.95,
                    bottom = noneiv_metric_mean - k*  noneiv_bar_size,
                    color=colors[1],
                    alpha=0.5)

plt.ylim(bottom=0, top=ymax)
ax = plt.gca()
ax.set_xticks(np.arange(1,len(data_list)+1))
ax.set_xticklabels(data_list, rotation='vertical')
plt.axhline(0.95,0.0,1.0,color='k', linestyle='dashed')
plt.ylabel('coverage (total unc.)')
plt.tight_layout()
plt.savefig('results/figures/total_coverage_bar_plot.pdf')
